/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */
package sga;

import java.util.Random;

/**
 *
 * @author Frantisek Hruska, Jiri Kubalik
 */
public class Individual {
    
    /**
     * Genes of individual binary represented .
     */
    public int[] genes;
    
    /**
     * Fitness of individual.
     */
    public double fitness;
    
    /**
     * Constructor. Generates random individual.
     */
    public Individual() {
        
        // Allocate memory for genes array.
        this.genes = new int[SGA.maxGenes];
        
        // Randomly initialise genes of the individual.
        for(int i = 0; i < SGA.maxGenes; i++) {
            if(Math.random() < 0.5) 
                this.genes[i] = 0;
            else
                this.genes[i] = 1;
        }

        // Super-bad initialization
        for(int i = SGA.maxGenes/3; i < SGA.maxGenes; i++) {
            this.genes[i] = 0;
        }
    }
    
    /**
     * Constructor generating empty individual without its genes defined.
     * @param b dump parameter.
     */
    public Individual(boolean b) {
        
        // Only allocate memory for genes.
        this.genes = new int[SGA.maxGenes];
    }
    
    /**
     * Cloning method for creating deep copies of individual instances.
     * @return 
     */
    @Override
    public Individual clone() {
        
        Individual newInd = new Individual(false); // Create new empty idividual
        for(int i = 0; i < SGA.maxGenes; i++) {
            newInd.genes[i] = this.genes[i];
        }
        
        newInd.fitness = this.fitness;
        
        return newInd;
    }

    /**
     * Fitness function calculation.
     */
    public void getFitness(){
        double pom = Math.pow(2.0, SGA.maxGenes) - 1.0;
        double x;

        // 1. Decode parameter x
        x = this.decode(0, SGA.maxGenes);      // get an integer value that is encoded in the vector 'genes'
        x = SGA.minX + (x / pom) * SGA.rangeX;     // get a real value in the range <SGA.minX, SGA.minX + SGA.rangeX)

        // 2. Calculate f(x)
        this.fitness = calculate_fitness(x);
    }
    
    /**
    * Decodes genotype stored in 'genes'.
    * @param from place from which will be chromosome decoded
    * @param to place where decoding will be stopped
    * @return an integer value coded by binary genes
    */
    public double decode(int from, int to){
	int power = 1;
	double temp = 0.0;

        for(int i = from; i < to; i++){
            temp += (power * this.genes[i]);
            power *= 2;
        }
        
        return temp;
    }

   /** 
    * Fitness Function Definition - highly multimodal function.
    * 1. f(x) = 2.0 + (x / 50.0) + Math.sin(x) + 2.0*Math.sin(Math.PI * x / 5.0), x in (0, 150) 
    * @param x 
    */
    public double calculate_fitness(double x){
        
        return 2.0 + (x / 50.0) + Math.sin(x) + 2.0*Math.sin(Math.PI * x / 50.0);
    }
    
    /**
     * One point crossover.
     * @param parent second individual in crossover
     * @return pair of newly generated individuals
     */
    public Individual[] one_point_crossover(Individual parent) {
        Random rnd = new Random();
        Individual[] offspring = new Individual[2];
        
        // Make clones of this and input individual.
        offspring[0] = this.clone();
        offspring[1] = parent.clone();
       
        // Generate crosspoint.
        int cross_point = 1 + rnd.nextInt(SGA.maxGenes - 1);
        
        // Switch first part of chromozomes.
        for(int i = 0; i <= cross_point; i++) {
            offspring[0].genes[i] = parent.genes[i];
            offspring[1].genes[i] = this.genes[i];
        }
        // Copy rest of chromozomes.
        for(int i = cross_point + 1; i < SGA.maxGenes; i++) {
            offspring[0].genes[i] = this.genes[i];
            offspring[1].genes[i] = parent.genes[i];
        }
        
        return offspring;
    }
    
    /**
     * One point crossover.
     * @param parent second individual in crossover
     * @return newly generated individuals
     */
    public Individual[] two_point_crossover(Individual parent) {
        Random rnd = new Random();
        Individual[] offspring = new Individual[2];

        // Make clones of this and input individual.
        offspring[0] = this.clone();
        offspring[1] = parent.clone();

        // Generate crosspoints.
        int crossPoint1 = rnd.nextInt(SGA.maxGenes);
        int crossPoint2 = rnd.nextInt(SGA.maxGenes);
        // While both cross points are the same, generate new cross point.
        while(crossPoint1 == crossPoint2){
            crossPoint2 = rnd.nextInt(SGA.maxGenes);
        }
        // Make sutre that the crossPoint2 is larger than crossPoint1.
        if(crossPoint2 < crossPoint1) {
            int temp = crossPoint1;
            crossPoint1 = crossPoint2;
            crossPoint2 = temp;
        }
        
        // Crossing method.
        // Copy first part of chromozomes.
        for(int i = 0; i < crossPoint1; i++) {
            offspring[0].genes[i] = this.genes[i];
            offspring[1].genes[i] = parent.genes[i];
        }
        // Switch middle parts of chromozomes.
        for(int i = crossPoint1; i < crossPoint2; i++) {
            offspring[0].genes[i] = parent.genes[i];
            offspring[1].genes[i] =this.genes[i];
        }
        // Copy rest of chromozomes.
        for(int i = crossPoint2; i < SGA.maxGenes; i++) {
            offspring[0].genes[i] = this.genes[i];
            offspring[1].genes[i] = parent.genes[i];
        }
        
        return offspring;
    }
    
    /**
     * Uniform crossover. Each gene is switched under some probability value.
     * @param id1 index of first individual
     * @param id2 index of second individual
     */
    public Individual[] uniform_crossover(Individual parent) {
        Random rnd = new Random();
        Individual[] offspring = new Individual[2];

        // Make clones of this and input individual.
        offspring[0] = this.clone();
        offspring[1] = parent.clone();
        
        // Each gene is inherited either from the 1st or the 2nd parent
        for(int geneIter = 0; geneIter < SGA.maxGenes; geneIter++) {
            if(rnd.nextBoolean()) 
                offspring[0].genes[geneIter] = parent.genes[geneIter];
            else 
                offspring[0].genes[geneIter] = this.genes[geneIter];
            
            if(rnd.nextBoolean()) 
                offspring[1].genes[geneIter] = this.genes[geneIter];
            else
                offspring[1].genes[geneIter] = parent.genes[geneIter];
        }
        
        return offspring;
    }
    
    /**
     * Find proper crossover according settings in SGA.java file.
     * @param id1 index of first individual
     * @param id2 index of second individual
     */
    public Individual[] crossover(Individual p) {
        switch(SGA.crossType) {
            case 0: return uniform_crossover(p);
            case 1: return one_point_crossover(p);
            case 2: return two_point_crossover(p);
            default:
                System.err.println("ERROR - Bad value of crossType variable in SGA.java file. Check if its value is between 0 and 2.");
                System.exit(1);
                return null;
        }
    }
    
    /**
     * Simple bit-flip mutation.
     */
    public void mutation() {
        Random rnd = new Random();

        for(int i = 0; i < SGA.maxGenes; i++) {
            // Mutate each gene with probability Pm.
            if(rnd.nextDouble() < SGA.pM) {
                this.genes[i] = (this.genes[i] + 1) % 2;    // swap between 0 and 1
            }
        }
    }
}
